from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import simulation_simulator_iverilog
from veriloggen import *

expected_verilog = """
module test #
  ( 
   parameter WIDTH = 8
  )
  (
  );

  reg CLK;
  reg RST;
  wire [WIDTH-1:0] LED;

  blinkled #
   (
    .WIDTH(WIDTH)
   )
  uut
   (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
   );

//  initial begin
//    $dumpfile("uut.vcd");
//    $dumpvars(0, uut);
//  end

  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end

  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    $finish;
  end

endmodule

module blinkled #
  ( 
   parameter WIDTH = 8
  )
  ( 
   input CLK, 
   input RST,
   output reg [WIDTH-1:0] LED
  );
  reg [32-1:0] count;
  always @(posedge CLK) begin
    if(RST) begin
      count <= 0;
    end else begin
      if(count == 15) begin
        count <= 0;
      end else begin
        count <= count + 1;
      end
    end
  end
  always @(posedge CLK) begin
    if(RST) begin
      LED <= 0;
    end else begin
      $display("LED:%d count:%d", LED, count);
      if(count == 15) begin
        LED <= LED + 1;
      end 
    end
  end
endmodule
"""

expected_rslt = """\
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  x count:         x
LED:  0 count:         0
LED:  0 count:         1
LED:  0 count:         2
LED:  0 count:         3
LED:  0 count:         4
LED:  0 count:         5
LED:  0 count:         6
LED:  0 count:         7
LED:  0 count:         8
LED:  0 count:         9
LED:  0 count:        10
LED:  0 count:        11
LED:  0 count:        12
LED:  0 count:        13
LED:  0 count:        14
LED:  0 count:        15
LED:  1 count:         0
LED:  1 count:         1
LED:  1 count:         2
LED:  1 count:         3
LED:  1 count:         4
LED:  1 count:         5
LED:  1 count:         6
LED:  1 count:         7
LED:  1 count:         8
LED:  1 count:         9
LED:  1 count:        10
LED:  1 count:        11
LED:  1 count:        12
LED:  1 count:        13
LED:  1 count:        14
LED:  1 count:        15
LED:  2 count:         0
LED:  2 count:         1
LED:  2 count:         2
LED:  2 count:         3
LED:  2 count:         4
LED:  2 count:         5
LED:  2 count:         6
LED:  2 count:         7
LED:  2 count:         8
LED:  2 count:         9
LED:  2 count:        10
LED:  2 count:        11
LED:  2 count:        12
LED:  2 count:        13
LED:  2 count:        14
LED:  2 count:        15
LED:  3 count:         0
LED:  3 count:         1
LED:  3 count:         2
LED:  3 count:         3
LED:  3 count:         4
LED:  3 count:         5
LED:  3 count:         6
LED:  3 count:         7
LED:  3 count:         8
LED:  3 count:         9
LED:  3 count:        10
LED:  3 count:        11
LED:  3 count:        12
LED:  3 count:        13
LED:  3 count:        14
LED:  3 count:        15
LED:  4 count:         0
LED:  4 count:         1
LED:  4 count:         2
LED:  4 count:         3
LED:  4 count:         4
LED:  4 count:         5
LED:  4 count:         6
LED:  4 count:         7
LED:  4 count:         8
LED:  4 count:         9
LED:  4 count:        10
LED:  4 count:        11
LED:  4 count:        12
LED:  4 count:        13
LED:  4 count:        14
LED:  4 count:        15
LED:  5 count:         0
LED:  5 count:         1
LED:  5 count:         2
LED:  5 count:         3
LED:  5 count:         4
LED:  5 count:         5
LED:  5 count:         6
LED:  5 count:         7
LED:  5 count:         8
LED:  5 count:         9
LED:  5 count:        10
LED:  5 count:        11
LED:  5 count:        12
LED:  5 count:        13
LED:  5 count:        14
LED:  5 count:        15
LED:  6 count:         0
LED:  6 count:         1
LED:  6 count:         2
LED:  6 count:         3
"""

def test():
    veriloggen.reset()
    test_module = simulation_simulator_iverilog.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)

    sim = simulation.Simulator(test_module, sim='iverilog')
    rslt = sim.run()
    
    assert(expected_rslt == rslt)



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes


_tmp_count = 0


def reset():
    global _tmp_count
    _tmp_count = 0


def _tmp_name(prefix='_tmp_fixed'):
    global _tmp_count
    v = _tmp_count
    _tmp_count += 1
    ret = '_'.join([prefix, str(v)])
    return ret


def FixedInput(m, name, width=32, point=0,
               length=None, signed=True):
    obj = _FixedInput(width=width, length=length, signed=signed,
                      name=name, module=m, point=point)
    m.add_object(obj)
    return obj


def FixedOutput(m, name, width=32, point=0,
                length=None, signed=True):
    obj = _FixedOutput(width=width, length=length, signed=signed,
                       name=name, module=m, point=point)
    m.add_object(obj)
    return obj


def FixedOutputReg(m, name, width=32, point=0,
                   length=None, signed=True, initval=None):
    obj = _FixedOutput(width=width, length=length, signed=signed,
                       name=name, module=m, point=point)
    m.add_object(obj)
    obj = _FixedReg(width=width, length=length, signed=signed, initval=initval,
                    name=name, module=m, point=point)
    m.add_object(obj)
    return obj


def FixedInout(m, name, width=32, point=0,
               length=None, signed=True):
    obj = _FixedInout(width=width, length=length, signed=signed,
                      name=name, module=m, point=point)
    m.add_object(obj)
    return obj


def FixedReg(m, name, width=32, point=0,
             length=None, signed=True, initval=None):
    obj = _FixedReg(width=width, length=length, signed=signed, initval=initval,
                    name=name, module=m, point=point)
    m.add_object(obj)
    return obj


def FixedWire(m, name, width=32, point=0,
              length=None, signed=True):
    obj = _FixedWire(width=width, length=length, signed=signed,
                     name=name, module=m, point=point)
    m.add_object(obj)
    return obj


def FixedTmpReg(m, width=32, point=0, length=None, signed=True, initval=None):
    name = _tmp_name()
    return FixedReg(m, name, width, point, length, signed, initval)


def FixedTmpWire(m, width=32, point=0, length=None, signed=True):
    name = _tmp_name()
    return FixedWire(m, name, width, point, length, signed)


def FixedConst(value, point=0, signed=True, raw=False):
    obj = _FixedInt(value, point=point, raw=raw, signed=signed)
    return obj


def reinterpret_cast_to_fixed(value, point, signed=True):
    m = value._get_module()
    width = value.bit_length()
    v = FixedTmpWire(m, width=width, point=point, signed=signed)
    v.assign_raw(value)
    return v


def to_fixed(value, point):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return value

    if isinstance(value, (int, bool, float)) and isinstance(point, int):
        mag = 2 ** point
        return int(value * mag)

    if isinstance(value, (int, bool)):
        mag = vtypes.Int(2) ** point
        return vtypes.Int(value) * mag

    if isinstance(value, float):
        mag = vtypes.Int(2) ** point
        return vtypes.Float(value) * mag

    signed = vtypes.get_signed(value)
    return shift_left(value, point, signed)


def fixed_to_int(value, point):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return value

    if isinstance(value, (int, bool, float)) and isinstance(point, int):
        mag = 2 ** point
        return value // mag

    if isinstance(value, (int, bool, float)):
        mag = vtypes.Int(2) ** point
        return vtypes.Int(value) // mag

    signed = vtypes.get_signed(value)
    return shift_right(value, point, signed)


def fixed_to_int_low(value, point):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return 0

    if isinstance(value, (int, bool, float)) and isinstance(point, int):
        mag = 2 ** point
        return int(value % mag)

    return vtypes.And(value, vtypes.Repeat(vtypes.Int(1, 1), point))


def fixed_to_real(value, point):
    if point < 0:
        raise ValueError('point must be more than 0')

    if point == 0:
        return vtypes.SystemTask('itor', value)

    if isinstance(value, float):
        raise TypeError("value is already float.")

    if isinstance(value, (int, bool)) and isinstance(point, int):
        mag = 2 ** point
        return float(value) / mag

    signed = vtypes.get_signed(value)

    width = value.bit_length()
    msb = (value[width - 1] if isinstance(value, vtypes._Variable) else
           (value >> (width - 1)) & 0x1)

    v0 = (vtypes.SystemTask('itor', fixed_to_int(value, point)) +
          vtypes.SystemTask('itor', fixed_to_int_low(value, point)) /
          vtypes.SystemTask('itor', vtypes.Int(2) ** point))

    nv = vtypes.Unot(value) + 1
    v1 = ((vtypes.SystemTask('itor', fixed_to_int(nv, point)) +
           vtypes.SystemTask('itor', fixed_to_int_low(nv, point)) /
           vtypes.SystemTask('itor', vtypes.Int(2) ** point))) * vtypes.SystemTask('itor', -1)

    return vtypes.Mux(signed and msb == 0, v0, v1)


def to_signed(value):
    if isinstance(value, int):
        return vtypes.Int(value, signed=True)
    if isinstance(value, vtypes.Int):
        return vtypes.Int(value.value, signed=True)
    if vtypes.get_signed(value):
        return value
    data = vtypes.SystemTask('signed', value)
    point = value.point if hasattr(value, 'point') else 0
    return _FixedSkipUnaryOperator(data, point, True)


def adjust(left, right, lpoint, rpoint, signed=True):
    diff_lpoint = 0 if rpoint < lpoint else rpoint - lpoint
    diff_rpoint = 0 if lpoint < rpoint else lpoint - rpoint
    ldata = left if diff_lpoint == 0 else shift_left(left, diff_lpoint, signed)
    rdata = right if diff_rpoint == 0 else shift_left(
        right, diff_rpoint, signed)
    _ldata = to_signed(ldata) if signed and diff_lpoint != 0 else ldata
    _rdata = to_signed(rdata) if signed and diff_rpoint != 0 else rdata
    return _ldata, _rdata


def write_adjust(value, point):
    lpoint = point
    rvalue = value
    rsigned = vtypes.get_signed(value)
    if not isinstance(value, _FixedBase):
        rpoint = 0
    else:
        rpoint = value.point

    ldiff = vtypes.Mux(lpoint <= rpoint, 0, lpoint - rpoint)
    rdiff = vtypes.Mux(lpoint >= rpoint, 0, rpoint - lpoint)
    v = vtypes.Mux(lpoint > rpoint, shift_left(rvalue, ldiff, rsigned),
                   vtypes.Mux(lpoint < rpoint, shift_right(rvalue, rdiff, rsigned), rvalue))
    return v


def shift_left(value, size, signed=True):
    if isinstance(value, vtypes.Int):
        value = value.value

    if isinstance(value, vtypes.Float):
        value = value.value

    if isinstance(value, int) and isinstance(size, int):
        return value << size

    if isinstance(value, bool) and isinstance(size, int):
        return value << size

    if isinstance(value, float) and isinstance(size, int):
        return value * (2 ** size)

    return vtypes.Sll(value, size)


def shift_right(value, size, signed=True):
    if isinstance(value, vtypes.Int):
        value = value.value

    if isinstance(value, vtypes.Float):
        value = value.value

    if isinstance(value, int) and isinstance(size, int):
        return value >> size

    if isinstance(value, bool) and isinstance(size, int):
        return value >> size

    if isinstance(value, float) and isinstance(size, int):
        return value / (2 ** size)

    if signed:
        return vtypes.Sra(value, size)

    return vtypes.Srl(value, size)


def _max_mux(a, b):
    return vtypes.Mux(a > b, a, b)


def _min_mux(a, b):
    return vtypes.Mux(a < b, a, b)


class _FixedBase(object):
    ast_name = None

    @property
    def int_part(self):
        return vtypes.Sra(self, self.point)

    @property
    def dec_part(self):
        mask = vtypes.Mux(self.point == 0, 0, vtypes.Repeat(
            vtypes.Int(1, width=1), self.point))
        return self & mask

    def to_int(self):
        return self.int_part

    def _adjust(self, value):
        return write_adjust(value, self.point)

    @property
    def raw(self):
        width = self.width if self.width is not None else 1
        return self[0:width]

    def __lt__(self, r):
        return _FixedLessThan(self, r)

    def __le__(self, r):
        return _FixedLessEq(self, r)

    def __eq__(self, r):
        return _FixedEq(self, r)

    def __ne__(self, r):
        return _FixedNotEq(self, r)

    def __ge__(self, r):
        return _FixedGreaterEq(self, r)

    def __gt__(self, r):
        return _FixedGreaterThan(self, r)

    def __add__(self, r):
        return _FixedPlus(self, r)

    def __sub__(self, r):
        return _FixedMinus(self, r)

    def __mul__(self, r):
        return _FixedTimes(self, r)

    def __div__(self, r):
        return _FixedDivide(self, r)

    def __truediv__(self, r):
        return _FixedDivide(self, r)

    def __lshift__(self, r):
        return _FixedSll(self, r)

    def __rshift__(self, r):
        return _FixedSra(self, r)

    def __neg__(self):
        return _FixedUminus(self)

    def __pos__(self):
        return _FixedUplus(self)


class _FixedVariable(_FixedBase, vtypes._Variable):
    ast_name = None
    __hash__ = vtypes._Variable.__hash__
    no_write_check = True

    def __init__(self, width=1, length=None, signed=True, value=None, initval=None, name=None,
                 module=None, point=0):
        self.point = point
        if initval is not None and not isinstance(initval, _FixedConstant):
            initval = to_fixed(initval, point)
        vtypes._Variable.__init__(
            self, width, length, signed, value, initval, name, module)

    def write(self, value, blk=False, ldelay=None, rdelay=None):
        v = self._adjust(value)
        return vtypes._Variable.write(self, v, blk=blk, ldelay=ldelay, rdelay=rdelay)

    def write_raw(self, value, blk=False, ldelay=None, rdelay=None):
        return vtypes._Variable.write(self, value, blk=blk, ldelay=ldelay, rdelay=rdelay)

    def assign(self, value):
        v = self._adjust(value)
        return vtypes._Variable.assign(self, v)

    def assign_raw(self, value):
        if self.module is None:
            raise ValueError(
                "Variable '%s' has no parent module information" % self.name)
        return self.module.Assign(self.write_raw(value))

    def connect(self, value):
        v = self._adjust(value)
        return vtypes._Variable.connect(self, v)


class _FixedInput(_FixedVariable, vtypes.Input):
    ast_name = 'Input'


class _FixedOutput(_FixedVariable, vtypes.Output):
    ast_name = 'Output'


class _FixedInout(_FixedVariable, vtypes.Inout):
    ast_name = 'Inout'


class _FixedReg(_FixedVariable, vtypes.Reg):
    ast_name = 'Reg'


class _FixedWire(_FixedVariable, vtypes.Wire):
    ast_name = 'Wire'


class _FixedBinaryOperator(_FixedBase, vtypes._BinaryOperator):
    ast_name = None
    overwrite_signed = False
    __hash__ = vtypes._BinaryOperator.__hash__

    def init(self, left, right):
        lpoint = left.point if isinstance(left, _FixedBase) else 0
        rpoint = right.point if isinstance(right, _FixedBase) else 0
        lsigned = vtypes.get_signed(left)
        rsigned = vtypes.get_signed(right)
        point = _max_mux(lpoint, rpoint)
        signed = lsigned and rsigned if not self.overwrite_signed else False
        ldata, rdata = adjust(left, right, lpoint, rpoint, signed)
        self.point = point
        return ldata, rdata


class _FixedBinaryShiftOperator(_FixedBinaryOperator):

    def init(self, left, right):
        lpoint = left.point if isinstance(left, _FixedBase) else 0
        rpoint = right.point if isinstance(right, _FixedBase) else 0
        if rpoint != 0:
            raise TypeError("shift amount must be int")
        point = lpoint
        ldata, rdata = left, right
        self.point = point
        return ldata, rdata


class _FixedUnaryOperator(_FixedBase, vtypes._UnaryOperator):
    ast_name = None
    __hash__ = vtypes._UnaryOperator.__hash__

    def __init__(self, right):
        vtypes._UnaryOperator.__init__(self, right)
        self.point = right.point if isinstance(right, _FixedBase) else 0


class _FixedSkipUnaryOperator(_FixedUnaryOperator, vtypes._SkipUnaryOperator):
    ast_name = '_SkipUnaryOperator'

    def __init__(self, right, point=None, signed=None):
        vtypes._SkipUnaryOperator.__init__(self, right)
        self.point = point if point is not None else 0
        self.signed = signed if signed is not None else True


class _FixedTimes(_FixedSkipUnaryOperator):
    ast_name = '_SkipUnaryOperator'
    overwrite_signed = False

    def __init__(self, left, right):
        lpoint = left.point if isinstance(left, _FixedBase) else 0
        rpoint = right.point if isinstance(right, _FixedBase) else 0
        lsigned = vtypes.get_signed(left)
        rsigned = vtypes.get_signed(right)
        point = _max_mux(lpoint, rpoint)
        signed = lsigned and rsigned if not self.overwrite_signed else False
        ldata = to_signed(left) if signed else left
        rdata = to_signed(right) if signed else right
        shift_size = _min_mux(lpoint, rpoint)
        data = vtypes.Times(ldata, rdata)
        if signed:
            data = vtypes.Sra(data, shift_size)
        else:
            data = vtypes.Srl(data, shift_size)
        _FixedSkipUnaryOperator.__init__(self, data, point, signed)


class _FixedDivide(_FixedSkipUnaryOperator):
    ast_name = '_SkipUnaryOperator'
    overwrite_signed = False

    def __init__(self, left, right):
        lpoint = left.point if isinstance(left, _FixedBase) else 0
        rpoint = right.point if isinstance(right, _FixedBase) else 0
        lsigned = vtypes.get_signed(left)
        rsigned = vtypes.get_signed(right)
        point = _max_mux(lpoint, rpoint)
        signed = lsigned and rsigned if not self.overwrite_signed else False
        lwidth = left.bit_length()
        rwidth = right.bit_length()

        if lpoint <= rpoint:
            ldata, rdata = adjust(left, right, lpoint, rpoint, signed)
            shift_size = point
        else:
            ldata = left
            rdata = right
            shift_size = point - (lpoint - rpoint)

        try:
            lmsb = ldata[lwidth - 1]
        except:
            lmsb = (ldata >> (lwidth - 1) & vtypes.Int(1, 1, base=2))

        try:
            rmsb = rdata[rwidth - 1]
        except:
            rmsb = (rdata >> (rwidth - 1) & vtypes.Int(1, 1, base=2))

        abs_ldata = (ldata if not lsigned else
                     vtypes.Mux(vtypes.Ulnot(lmsb), ldata, vtypes.Unot(ldata) + 1))
        abs_rdata = (rdata if not rsigned else
                     vtypes.Mux(vtypes.Ulnot(rmsb), rdata, vtypes.Unot(rdata) + 1))
        abs_data = vtypes.Divide(abs_ldata, abs_rdata)
        data = (abs_data if not signed else
                vtypes.Mux(vtypes.Eq(lmsb, rmsb),
                           abs_data, vtypes.Unot(abs_data) + 1))

        if shift_size > 0:
            data = vtypes.Sll(data, shift_size)

        _FixedSkipUnaryOperator.__init__(self, data, point, signed)


class _FixedPlus(_FixedBinaryOperator, vtypes.Plus):
    ast_name = 'Plus'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.Plus.__init__(self, left, right)


class _FixedMinus(_FixedBinaryOperator, vtypes.Minus):
    ast_name = 'Minus'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.Minus.__init__(self, left, right)


class _FixedSll(_FixedBinaryShiftOperator, vtypes.Sll):
    ast_name = 'Sll'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.Sll.__init__(self, left, right)


class _FixedSrl(_FixedBinaryShiftOperator, vtypes.Srl):
    ast_name = 'Srl'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.Srl.__init__(self, left, right)


class _FixedSra(_FixedBinaryShiftOperator, vtypes.Sra):
    ast_name = 'Sra'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.Sra.__init__(self, left, right)


class _FixedLessThan(_FixedBinaryOperator, vtypes.LessThan):
    ast_name = 'LessThan'
    overwrite_signed = True

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.LessThan.__init__(self, left, right)


class _FixedGreaterThan(_FixedBinaryOperator, vtypes.GreaterThan):
    ast_name = 'GreaterThan'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.GreaterThan.__init__(self, left, right)


class _FixedLessEq(_FixedBinaryOperator, vtypes.LessEq):
    ast_name = 'LessEq'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.LessEq.__init__(self, left, right)


class _FixedGreaterEq(_FixedBinaryOperator, vtypes.GreaterEq):
    ast_name = 'GreaterEq'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.GreaterEq.__init__(self, left, right)


class _FixedEq(_FixedBinaryOperator, vtypes.Eq):
    ast_name = 'Eq'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.Eq.__init__(self, left, right)


class _FixedNotEq(_FixedBinaryOperator, vtypes.NotEq):
    ast_name = 'NotEq'

    def __init__(self, left, right):
        left, right = self.init(left, right)
        vtypes.NotEq.__init__(self, left, right)


class _FixedUplus(_FixedUnaryOperator, vtypes.Uplus):
    ast_name = 'Uplus'
    pass


class _FixedUminus(_FixedUnaryOperator, vtypes.Uminus):
    ast_name = 'Uminus'
    pass


class _FixedConstant(_FixedBase):

    def __init__(self, orig_value=None):
        self.orig_value = orig_value


class _FixedInt(_FixedConstant, vtypes.Int):
    ast_name = 'Int'
    __hash__ = vtypes.Int.__hash__

    def __init__(self, value, width=None, base=None, point=0, signed=True, raw=False):
        value = value if raw else to_fixed(value, point)
        _FixedConstant.__init__(self, None)
        vtypes.Int.__init__(self, value, width, base, signed)
        self.point = point

    @property
    def raw(self):
        return vtypes.Int(self.value, self.width, self.base, self.signed)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import collections

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *

def mkLed():
    m = StubModule('blinkled')
    return m

def mkTop():
    m = Module('top')
    width = m.Parameter('WIDTH', 8)
    clk = m.Input('CLK')
    rst = m.Input('RST')
    led = m.Output('LED', width)
    
    params = ( width, )
    ports = ( clk, rst, led )
    
    # StubModule is automatically created inside
    m.Instance('blinkled', 'inst_blinkled', params, ports)

    return m

if __name__ == '__main__':
    top = mkTop()
    verilog = top.to_verilog()
    print(verilog)



from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import copy
from functools import reduce

import veriloggen.core.vtypes as vtypes
from veriloggen.seq.seq import Seq


def _and_vars(*vars):
    if not vars:
        return vtypes.Int(1)
    ret = None
    for var in vars:
        if var is None:
            continue
        if ret is None:
            ret = var
        else:
            ret = vtypes.AndList(ret, var)
    if ret is None:
        return vtypes.Int(1)
    return ret


def _connect_ready(m, var, ready):
    if var is None:
        return

    prev_assign = var._get_assign()
    if not prev_assign:
        m.Assign(var(ready))
    elif (isinstance(prev_assign.statement.right, vtypes.Int) and
          prev_assign.statement.right.value == 1):
        prev_assign.overwrite_right(ready)
        m.remove(prev_assign)
        m.append(prev_assign)
    else:
        prev_assign.overwrite_right(
            _and_vars(prev_assign.statement.right, ready))
        m.remove(prev_assign)
        m.append(prev_assign)


class Pipeline(vtypes.VeriloggenNode):
    """ Pipeline Generator """

    def __init__(self, m, name, clk, rst, width=32):
        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.width = width

        self.tmp_count = 0
        self.max_stage_id = 0
        self.vars = []

        self.seq = Seq(self.m, self.name, clk, rst, nohook=True)
        self.data_visitor = DataVisitor(self)

        self.done = False

    #-------------------------------------------------------------------------
    def input(self, data, valid=None, ready=None, width=None):
        if ready is not None and not isinstance(ready, (vtypes.Wire, vtypes.Output)):
            raise TypeError('ready port of PipelineVariable must be Wire., not %s' %
                            str(type(ready)))
        ret = _PipelineVariable(self, 0, data, valid, ready,
                                _PipelineInterface(data, valid, ready))
        self.vars.append(ret)
        return ret

    #-------------------------------------------------------------------------
    # self.__call__() calls this method
    def stage(self, data, initval=0, width=None, preg=None):
        if width is None:
            width = self.width
        stage_id, raw_data, raw_valid, raw_ready = self.data_visitor.visit(
            data)
        tmp_data, tmp_valid, tmp_ready = self._make_tmp(raw_data, raw_valid, raw_ready,
                                                        width, initval)
        next_stage_id = stage_id + 1 if stage_id is not None else None

        ret = _PipelineVariable(self, next_stage_id,
                                tmp_data, tmp_valid, tmp_ready, data)

        self.vars.append(ret)
        if isinstance(preg, _PipelineVariable):
            preg._add_preg(next_stage_id, ret)

        if next_stage_id is not None and next_stage_id > self.max_stage_id:
            self.max_stage_id = next_stage_id

        return ret

    #-------------------------------------------------------------------------
    # Accumulator
    def acc_and(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.And], data, width, initval, resetcond)

    def acc_nand(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.And, vtypes.Unot], data, width, initval, resetcond)

    def acc_or(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Or], data, width, initval, resetcond)

    def acc_xor(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Xor], data, width, initval, resetcond)

    def acc_xnor(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Xor, vtypes.Unot], data, width, initval, resetcond)

    def acc_nor(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Or, vtypes.Unot], data, width, initval, resetcond)

    def acc_add(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Plus], data, width, initval, resetcond)

    def acc_sub(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Minus], data, width, initval, resetcond)

    def acc_mul(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Times], data, width, initval, resetcond)

    def acc_div(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Divide], data, width, initval, resetcond)

    def acc_mod(self, data, initval=0, resetcond=None, width=None):
        return self._accumulate([vtypes.Mod], data, width, initval, resetcond)

    def acc_max(self, data, initval=0, resetcond=None, width=None):
        def op(left, right):
            return vtypes.Cond(left > right, left, right)
        return self._accumulate([op], data, width, initval, resetcond, 'max')

    def acc_min(self, data, initval=0, resetcond=None, width=None):
        def op(left, right):
            return vtypes.Cond(left < right, left, right)
        return self._accumulate([op], data, width, initval, resetcond, 'min')

    def acc_custom(self, data, ops, initval=0, resetcond=None, width=None, label=None):
        if not isinstance(ops, (tuple, list)):
            ops = [ops]
        return self._accumulate(ops, data, width, initval, resetcond, label)

    #-------------------------------------------------------------------------
    def make_always(self, reset=(), body=()):
        if self.done:
            raise ValueError('make_always() has been already called.')
        self.done = True

        for var in self.vars:
            if var.output_vars is not None:
                var.make_output()

        self.m.Always(vtypes.Posedge(self.clk))(
            vtypes.If(self.rst)(
                reset,
                self.make_reset()
            )(
                body,
                self.make_code()
            ))

    #-------------------------------------------------------------------------
    def make_reset(self):
        return self.seq.make_reset()

    #-------------------------------------------------------------------------
    def make_code(self):
        return self.seq.make_code()

    #-------------------------------------------------------------------------
    def draw_graph(self, filename='out.png', prog='dot'):
        _draw_graph(self, filename, prog)

    #-------------------------------------------------------------------------
    def _accumulate(self, ops, data, width=None, initval=0, resetcond=None, oplabel=None):
        if width is None:
            width = self.width
        stage_id, raw_data, raw_valid, raw_ready = self.data_visitor.visit(
            data)
        tmp_data, tmp_valid, tmp_ready = self._make_tmp(raw_data, raw_valid, raw_ready,
                                                        width, initval, acc_ops=ops)
        next_stage_id = stage_id + 1 if stage_id is not None else None

        ret = _PipelineVariable(self, next_stage_id, tmp_data, tmp_valid, tmp_ready,
                                data, ops, resetcond, initval, oplabel)

        if resetcond is not None:
            ret.reset(resetcond, initval)

        self.vars.append(ret)

        return ret

    #-------------------------------------------------------------------------
    def _add_reg(self, prefix, count, width=None, initval=0):
        tmp_name = '_'.join(['', self.name, prefix, str(count)])
        tmp = self.m.Reg(tmp_name, width, initval=initval)
        return tmp

    def _add_wire(self, prefix, count, width=None):
        tmp_name = '_'.join(['', self.name, prefix, str(count)])
        tmp = self.m.Wire(tmp_name, width)
        return tmp

    #-------------------------------------------------------------------------
    def _make_tmp(self, data, valid, ready, width=None, initval=0, acc_ops=()):
        tmp_data = self._add_reg(
            'data', self.tmp_count, width=width, initval=initval)

        if valid is not None:
            tmp_valid = self._add_reg('valid', self.tmp_count, initval=0)
        else:
            tmp_valid = None

        if ready:
            tmp_ready = self._add_wire('ready', self.tmp_count)
        else:
            tmp_ready = None

        self.tmp_count += 1

        # all ready
        all_ready = None
        for r in ready:
            if r is None:
                continue
            if all_ready is None:
                all_ready = r
            else:
                all_ready = vtypes.AndList(all_ready, r)

        # data
        data_cond_vars = []
        if valid is not None:
            data_cond_vars.append(valid)
        if tmp_ready is not None:
            data_cond_vars.append(all_ready)
            if tmp_valid is not None:
                data_cond_vars.append(vtypes.OrList(
                    tmp_ready, vtypes.Not(tmp_valid)))
            else:
                data_cond_vars.append(tmp_ready)

        if len(data_cond_vars) == 0:
            data_cond = None
        elif len(data_cond_vars) == 1:
            data_cond = data_cond_vars[0]
        else:
            data_cond = vtypes.AndList(*data_cond_vars)

        # Accumulator
        for op in acc_ops:
            if not isinstance(op, type):
                data = op(tmp_data, data)
            elif issubclass(op, vtypes._BinaryOperator):
                data = op(tmp_data, data)
            elif issubclass(op, vtypes._UnaryOperator):
                data = op(data)

            if not isinstance(data, vtypes._Numeric):
                raise TypeError("Operator '%s' returns unsupported object type '%s'."
                                % (str(op), str(type(data))))

        self.seq.add(tmp_data(data), cond=data_cond)

        # valid
        valid_cond_vars = []
        if tmp_ready is not None:
            valid_cond_vars.append(all_ready)
            ordy = vtypes.OrList(tmp_ready, vtypes.Not(tmp_valid))
            valid_cond_vars.append(ordy)

        if len(valid_cond_vars) == 0:
            valid_cond = None
        elif len(valid_cond_vars) == 1:
            valid_cond = valid_cond_vars[0]
        else:
            valid_cond = vtypes.AndList(*valid_cond_vars)

        if tmp_valid is not None:
            if tmp_ready is not None:
                self.seq.add(tmp_valid(0), cond=vtypes.AndList(
                    tmp_valid, tmp_ready))
            self.seq.add(tmp_valid(valid), cond=valid_cond)

        # ready
        if tmp_ready is not None:
            ordy = vtypes.AndList(vtypes.OrList(
                tmp_ready, vtypes.Not(tmp_valid)), valid)
            for r in ready:
                _connect_ready(self.m, r, ordy)

        return tmp_data, tmp_valid, tmp_ready

    #-------------------------------------------------------------------------
    def _make_prev(self, data, valid, ready, width=None, initval=0):
        tmp_data = self._add_reg(
            'data', self.tmp_count, width=width, initval=initval)
        tmp_valid = valid
        tmp_ready = ready

        self.tmp_count += 1

        if valid is not None and ready is not None:
            data_cond = vtypes.AndList(valid, ready)
        elif valid is not None:
            data_cond = valid
        elif ready is not None:
            data_cond = ready
        else:
            data_cond = None

        self.seq.add(tmp_data(data), cond=data_cond)

        return tmp_data, tmp_valid, tmp_ready

    #-------------------------------------------------------------------------
    def __call__(self, data, initval=0, width=None):
        return self.stage(data, initval=initval, width=width)

#-------------------------------------------------------------------------


class _PipelineInterface(object):

    def __init__(self, data, valid=None, ready=None, output=False):
        self.data = data
        self.valid = valid
        self.ready = ready
        self.output = output

    def __str__(self):
        args = [self.data, self.valid, self.ready]
        return ','.join([str(arg) for arg in args])

#-------------------------------------------------------------------------


class _PipelineNumeric(vtypes._Numeric):
    pass


class _PipelineVariable(_PipelineNumeric):

    def __init__(self, df, stage_id, data, valid=None, ready=None,
                 src_data=None, ops=None, resetcond=None, initval=None, oplabel=None):
        self.df = df
        self.stage_id = stage_id
        self.data = data
        self.valid = valid
        self.ready = ready
        self.output_vars = None
        self.src_data = src_data
        self.dst_data = None
        self.ops = ops
        self.resetcond = resetcond
        self.initval = initval
        self.oplabel = oplabel
        self.prev_dict = {}
        self.preg_dict = {}

        if self.ready is not None:
            ready = vtypes.Int(1)
            _connect_ready(self.df.m, self.ready, ready)

    def prev(self, index, initval=0):
        if index == 0:
            return self
        if index < 0:
            raise ValueError('Index to a previous value must be positive.')

        if index in self.prev_dict:
            return self.prev_dict[index]

        width = self.data.bit_length()
        p = self

        for i in range(index):
            if (i + 1) in self.prev_dict:
                p = self.prev_dict[i + 1]
                continue

            tmp_data, tmp_valid, tmp_ready = self.df._make_prev(p.data, p.valid, p.ready,
                                                                width, initval)
            p = _PipelineVariable(self.df, p.stage_id,
                                  tmp_data, tmp_valid, tmp_ready, p)
            self.df.vars.append(p)
            self.prev_dict[i + 1] = p

        return p

    def output(self, data, valid=None, ready=None):
        self.output_vars = (data, valid, ready)

    def make_output(self):
        data, valid, ready = self.output_vars

        # Inserting delayed registers
        ovar = self
        if self.stage_id is not None:
            for i in range(self.stage_id, self.df.max_stage_id):
                ovar = self.df.stage(ovar, preg=ovar)

        if not isinstance(data, (vtypes.Wire, vtypes.Output)):
            raise TypeError('Data signal must be Wire, not %s' %
                            str(type(data)))
        else:
            ovar.df.m.Assign(data(ovar.data))

        my_valid = vtypes.Int(1) if ovar.valid is None else ovar.valid
        if valid is None:
            pass
        elif not isinstance(valid, (vtypes.Wire, vtypes.Output)):
            raise TypeError('Valid signal must be Wire, not %s' %
                            str(type(valid)))
        else:
            ovar.df.m.Assign(valid(my_valid))

        if ready is None:
            ready = vtypes.Int(1)

        _connect_ready(ovar.df.m, ovar.ready, ready)

        ovar.dst_data = _PipelineInterface(data, valid, ready, output=True)

    def reset(self, cond, initval=0):
        self.resetcond = cond
        self.initval = initval
        self.df.seq.add(self.data(initval), cond=cond)
        if self.valid is not None:
            self.df.seq.add(self.valid(0), cond=cond)

    def bit_length(self):
        return self.data.bit_length()

    def _add_preg(self, stage_id, var):
        self.preg_dict[stage_id] = var

    def _get_preg(self, stage_id):
        if stage_id is None:
            return self
        if self.stage_id is None:
            return self
        if stage_id == self.stage_id:
            return self
        return self.preg_dict[stage_id]

#-------------------------------------------------------------------------


class _PipelineVisitor(object):

    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))

    def visit(self, node):
        if isinstance(node, vtypes._BinaryOperator):
            return self.visit__BinaryOperator(node)

        if isinstance(node, vtypes._UnaryOperator):
            return self.visit__UnaryOperator(node)

        if isinstance(node, vtypes._Variable):
            return self.visit__Variable(node)

        if isinstance(node, vtypes._Constant):
            return self.visit__Constant(node)

        visitor = getattr(
            self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)

    def visit__PipelineVariable(self, node):
        raise NotImplementedError(
            'visit__PipelineVariable() must be implemented')

    def visit__BinaryOperator(self, node):
        raise NotImplementedError(
            'visit__BinaryOperator() must be implemented')

    def visit__UnaryOperator(self, node):
        raise NotImplementedError('visit__UnaryOperator() must be implemented')

    def visit_Pointer(self, node):
        raise NotImplementedError('visit_Pointer() must be implemented')

    def visit_Slice(self, node):
        raise NotImplementedError('visit_Slice() must be implemented')

    def visit_Cat(self, node):
        raise NotImplementedError('visit_Cat() must be implemented')

    def visit_Repeat(self, node):
        raise NotImplementedError('visit_Repeat() must be implemented')

    def visit_Cond(self, node):
        raise NotImplementedError('visit_Cond() must be implemented')

    def visit__Variable(self, node):
        raise NotImplementedError('visit__Variable() must be implemented')

    def visit__Constant(self, node):
        raise NotImplementedError('visit__Constant() must be implemented')

    def visit_bool(self, node):
        raise NotImplementedError('visit__Constant() must be implemented')

    def visit_int(self, node):
        raise NotImplementedError('visit__Constant() must be implemented')

    def visit_str(self, node):
        raise NotImplementedError('visit__Constant() must be implemented')

    def visit_float(self, node):
        raise NotImplementedError('visit__Constant() must be implemented')

#-------------------------------------------------------------------------


class DataVisitor(_PipelineVisitor):

    def __init__(self, df):
        self.df = df

    def pack_valid(self, lvalid, rvalid):
        if rvalid is not None and lvalid is not None:
            return vtypes.AndList(lvalid, rvalid)
        elif rvalid is None and lvalid is None:
            return None
        elif rvalid is None:
            return lvalid
        elif lvalid is None:
            return rvalid
        return None

    def pack_ready(self, lready, rready):
        return lready + rready

    def pack(self, *args):
        rslts = [self.visit(arg) for arg in args]
        all_stage_none = reduce(lambda t, x: t and x is None, [
                                rslt[0] for rslt in rslts], True)

        if all_stage_none:
            data = [rslt[1] for rslt in rslts]
            valid = reduce(self.pack_valid, [rslt[2] for rslt in rslts], None)
            ready = reduce(self.pack_ready, [rslt[3] for rslt in rslts], [])
            return (None, data, valid, ready)

        max_stage = reduce(lambda x, y:
                           None if x is None and y is None else
                           x if x is not None and y is None else
                           y if y is not None and x is None else
                           max(x, y), [rslt[0] for rslt in rslts], None)

        new_args = []
        for rslt, arg in zip(rslts, args):
            stage = rslt[0]
            if stage is None:
                new_args.append(arg)
                continue
            if stage == max_stage:
                new_args.append(arg)
                continue

            diff = max_stage - stage
            p = arg
            for i in range(diff):
                width = rslt[1].bit_length()
                p = self.df.stage(p, width=width, preg=arg)

            new_args.append(p)

        new_rslts = [self.visit(arg) for arg in new_args]
        data = [rslt[1] for rslt in new_rslts]
        valid = reduce(self.pack_valid, [rslt[2] for rslt in new_rslts], None)
        ready = reduce(self.pack_ready, [rslt[3] for rslt in new_rslts], [])

        return (max_stage, data, valid, ready)

    def make_valid(self, valid, ready):
        if ready is not None and valid is not None:
            next_valid = vtypes.AndList(valid, ready)
        elif ready is not None:
            next_valid = ready
        elif valid is not None:
            next_valid = valid
        else:
            next_valid = None
        return next_valid

    def visit__PipelineVariable(self, node):
        ready = [] if node.ready is None else [node.ready]
        #valid = self.make_valid(node.valid, node.ready)
        valid = node.valid
        return (node.stage_id, node.data, valid, ready)

    def visit__Variable(self, node):
        return (None, node, None, [])

    def visit__Constant(self, node):
        return (None, node, None, [])

    def visit__BinaryOperator(self, node):
        stage, data, valid, ready = self.pack(node.left, node.right)
        cls = type(node)
        return stage, cls(*data), valid, ready

    def visit__UnaryOperator(self, node):
        return self.visit(node.right)

    def visit_Pointer(self, node):
        stage, data, valid, ready = self.pack(node.var, node.pos)
        cls = type(node)
        return stage, cls(*data), valid, ready

    def visit_Slice(self, node):
        stage, data, valid, ready = self.pack(node.var, node.msb, node.lsb)
        cls = type(node)
        return stage, cls(*data), valid, ready

    def visit_Cat(self, node):
        stage, data, valid, ready = self.pack(*node.vars)
        cls = type(node)
        return stage, cls(*data), valid, ready

    def visit_Repeat(self, node):
        stage, data, valid, ready = self.pack(node.var, node.times)
        cls = type(node)
        return stage, cls(*data), valid, ready

    def visit_Cond(self, node):
        stage, data, valid, ready = self.pack(
            node.condition, node.true_value, node.false_value)
        cls = type(node)
        return stage, cls(*data), valid, ready

    def visit_bool(self, node):
        if node:
            return (None, vtypes.Int(1), None, [])
        return (None, vtypes.Int(0), None, [])

    def visit_int(self, node):
        return (None, vtypes.Int(node), None, [])

    def visit_str(self, node):
        return (None, vtypes.Str(node), None, [])

    def visit_float(self, node):
        return (None, vtypes.Float(node), None, [])

#-------------------------------------------------------------------------


def _draw_graph(df, filename='out.png', prog='dot'):
    gg = GraphGenerator(df)
    gg.draw(filename, prog)


class GraphGenerator(_PipelineVisitor):

    def __init__(self, df):
        try:
            import pygraphviz as pgv
        except:
            raise ImportError(
                'Graph generator of lib.Pipeline requires Pygraphviz.')
        self.df = df
        self.graph = pgv.AGraph(directed=True)
        self.visited_node = set()
        self.const_node = []

    def draw(self, filename='out.png', prog='dot'):
        for var in self.df.vars:
            # traverse from only output nodes
            if isinstance(var.dst_data, _PipelineInterface):
                self.visit(var)

        self.graph.write('out.dot')
        self.graph.layout(prog=prog)
        self.graph.draw(filename)

    def _add_node(self, node, label=None, color='black', shape='box', style='solid'):
        self.visited_node.add(id(node))
        if label is None:
            self.graph.add_node(id(node), color=color,
                                shape=shape, style=style)
        else:
            self.graph.add_node(id(node), label=label,
                                color=color, shape=shape, style=style)

    def _add_edge(self, start, end, color='black', label=None, style='solid'):
        if isinstance(start, (bool, int, str, float)):
            # to append multiple first-class objects with same values
            start = copy.deepcopy(vtypes._Constant(start))
            self._add_node(start, label=str(start), shape='invtriangle')
            self.const_node.append(start)

        if isinstance(end, (bool, int, str, float)):
            # to append multiple first-class objects with same values
            end = copy.deepcopy(vtypes._Constant(end))
            self._add_node(end, label=str(end), shape='invtriangle')
            self.const_node.append(end)

        if label:
            self.graph.add_edge(id(start), id(
                end), color=color, label=label, style=style)
        else:
            self.graph.add_edge(id(start), id(end), color=color, style=style)

    def _max_stage_id(self, *args):
        maxval = None
        for arg in args:
            if arg is None or not hasattr(arg, 'stage_id'):
                continue
            if arg.stage_id is None:
                continue
            if maxval is None:
                maxval = arg.stage_id
                continue
            if arg.stage_id > maxval:
                maxval = arg.stage_id
        return maxval

    def _visited(self, node):
        return id(node) in self.visited_node

    def visit(self, node):
        if self._visited(node):
            return None
        return _PipelineVisitor.visit(self, node)

    def visit__PipelineInterface(self, node):
        if node.output:
            self._add_node(node, label=str(node), shape='trapezium')
        else:
            self._add_node(node, label=str(node), shape='invtrapezium')

    def visit__PipelineVariable(self, node):
        # standard
        if node.src_data is not None and not node.ops:
            if node.stage_id is None:
                self._add_node(node, label=str(node.data),
                               shape='box', style='dashed')
            else:
                self._add_node(node, label=''.join(
                    [str(node.stage_id), ':', str(node.data)]), shape='box')
            self.visit(node.src_data)
            self._add_edge(node.src_data, node)
        # accumulator
        if node.src_data is not None and node.ops:
            label = [str(node.stage_id)]
            label.append(':')
            label.append(str(node.data))
            if node.oplabel is not None:
                label.append(' ')
                label.append(node.oplabel)
            else:
                for op in node.ops:
                    if isinstance(op, type):
                        label.append(' ')
                        label.append(vtypes.op2mark(op.__name__))
                    else:
                        label.append(' C')
            label.append('=')
            self._add_node(node, label=''.join(label),
                           shape='box', style='rounded')
            self.visit(node.src_data)
            self._add_edge(node.src_data, node)
            if node.resetcond:
                self.visit(node.resetcond)
                self._add_edge(node.resetcond, node,
                               label='RST', style='dashed')
        if isinstance(node.dst_data, _PipelineInterface):
            self.visit(node.dst_data)
            self._add_edge(node, node.dst_data)

    def visit__Variable(self, node):
        self._add_node(node, label=node.name, shape='invtriangle')

    def visit__Constant(self, node):
        self._add_node(node, label=node.value, shape='invtriangle')

    def visit__BinaryOperator(self, node):
        mark = vtypes.op2mark(node.__class__.__name__)
        self._add_node(node, label=mark, shape='ellipse')
        maxid = self._max_stage_id(node.left, node.right)
        left = node.left
        right = node.right
        if hasattr(left, '_get_preg'):
            left = left._get_preg(maxid)
        if hasattr(right, '_get_preg'):
            right = right._get_preg(maxid)
        self.visit(left)
        self.visit(right)
        self._add_edge(left, node, label='L')
        self._add_edge(right, node, label='R')

    def visit__UnaryOperator(self, node):
        mark = vtypes.op2mark(node.__class__.__name__)
        self._add_node(node, label=mark, shape='ellipse')
        self.visit(node.right)
        self._add_edge(node.right, node)

    def visit_Pointer(self, node):
        mark = 'sel'
        self._add_node(node, label=mark, shape='ellipse')
        maxid = self._max_stage_id(node.var, node.pos)
        var = node.var
        pos = node.pos
        if hasattr(var, '_get_preg'):
            var = var._get_preg(maxid)
        if hasattr(pos, '_get_preg'):
            pos = pos._get_preg(maxid)
        self.visit(var)
        self.visit(pos)
        self._add_edge(var, node, label='V')
        self._add_edge(pos, node, label='P', style='dashed')

    def visit_Slice(self, node):
        mark = 'slice'
        self._add_node(node, label=mark, shape='ellipse')
        maxid = self._max_stage_id(node.var, node.msb, node.lsb)
        var = node.var
        msb = node.msb
        lsb = node.lsb
        if hasattr(var, '_get_preg'):
            var = var._get_preg(maxid)
        if hasattr(msb, '_get_preg'):
            msb = msb._get_preg(maxid)
        if hasattr(lsb, '_get_preg'):
            lsb = lsb._get_preg(maxid)
        self.visit(var)
        self.visit(msb)
        self.visit(lsb)
        self._add_edge(var, node, label='V')
        self._add_edge(msb, node, label='M', style='dashed')
        self._add_edge(lsb, node, label='L', style='dashed')

    def visit_Cat(self, node):
        mark = 'cat'
        self._add_node(node, label=mark, shape='ellipse')
        maxid = self._max_stage_id(*node.vars)
        for var in node.vars:
            if hasattr(var, '_get_preg'):
                var = var._get_preg(maxid)
            self.visit(var)
            self._add_edge(var, node)

    def visit_Repeat(self, node):
        mark = 'repeat'
        self._add_node(node, label=mark, shape='ellipse')
        maxid = self._max_stage_id(node.var, node.times)
        var = node.var
        times = node.times
        if hasattr(var, '_get_preg'):
            var = var._get_preg(maxid)
        if hasattr(times, '_get_preg'):
            times = times._get_preg(maxid)
        self.visit(var)
        self.visit(times)
        self._add_edge(var, node, label='V')
        self._add_edge(times, node, label='T', style='dashed')

    def visit_Cond(self, node):
        mark = 'cond'
        self._add_node(node, label=mark, shape='ellipse')
        maxid = self._max_stage_id(
            node.condition, node.true_value, node.false_value)
        condition = node.condition
        true_value = node.true_value
        false_value = node.false_value
        if hasattr(condition, '_get_preg'):
            condition = condition._get_preg(maxid)
        if hasattr(true_value, '_get_preg'):
            true_value = true_value._get_preg(maxid)
        if hasattr(false_value, '_get_preg'):
            false_value = false_value._get_preg(maxid)
        self.visit(condition)
        self.visit(true_value)
        self.visit(false_value)
        self._add_edge(condition, node, label='C', style='dashed')
        self._add_edge(true_value, node, label='T')
        self._add_edge(false_value, node, label='F')

    def visit_bool(self, node):
        pass
        #self._add_node(node, label=str(node), shape='invtriangle')

    def visit_int(self, node):
        pass
        #self._add_node(node, label=str(node), shape='invtriangle')

    def visit_str(self, node):
        pass
        #self._add_node(node, label=str(node), shape='invtriangle')

    def visit_float(self, node):
        pass
        #self._add_node(node, label=str(node), shape='invtriangle')



from __future__ import absolute_import
from __future__ import print_function

import functools
import inspect
from collections import OrderedDict

import veriloggen.core.vtypes as vtypes
import veriloggen.types.util as util
from veriloggen.seq.seq import Seq


__intrinsics__ = ('intrinsic', 'statement', 'subst',
                  'display', 'write', 'finish', 'signed',
                  'embedded_code', 'embedded_numeric',
                  'set_parallel', 'unset_parallel')


def intrinsic(fsm, func, *args, **kwargs):
    """ function call as an intrinsic """
    func(fsm, *args, **kwargs)


class _verilog_meta(type):
    """ metaclass for verilog operator intrinsics """

    __verilog_classes__ = dict(inspect.getmembers(vtypes))
    __intrinsics__ = tuple(__verilog_classes__.keys())

    def __getattr__(self, key):
        if key in self.__verilog_classes__:
            cls = self.__verilog_classes__[key]

            @functools.wraps(cls)
            def wrapper(fsm, *args, **kwargs):
                return cls(*args, **kwargs)
            return wrapper

        raise NameError("name '%s' is not defined" % key)


_verilog = _verilog_meta('verilog', (object,),
                         {'__doc__': _verilog_meta.__doc__})


class verilog(_verilog):
    """ verilog operator intrinsics """
    pass


# for SingleStatement and EmbeddedCode
def statement(fsm, *values):
    for value in values:
        fsm(
            value
        )
    fsm.goto_next()


def subst(fsm, dst, src):
    statement(fsm, vtypes.Subst(dst, src))


def display(fsm, *args):
    statement(fsm, vtypes.Display(*args))


def write(fsm, *args):
    statement(fsm, vtypes.Write(*args))


def finish(fsm):
    statement(fsm, vtypes.Finish())


def signed(fsm, value):
    return vtypes.Signed(value)


def embedded_code(fsm, *codes):
    codes = [code.value if isinstance(code, vtypes.Str) else code
             for code in codes]
    code = '\n'.join(codes)
    statement(fsm, vtypes.EmbeddedCode(code))


def embedded_numeric(fsm, code):
    return vtypes.EmbeddedNumeric(code)


# parallel subst
def set_parallel(fsm):
    fsm.parallel = True


def unset_parallel(fsm):
    fsm.parallel = False
    fsm.goto_next()


def Lock(m, name, clk, rst, width=32):
    """ alias of Mutex class """
    return Mutex(m, name, clk, rst, width)


class Mutex(object):
    __intrinsics__ = ('lock', 'try_lock', 'unlock',
                      'acquire', 'release')

    def __init__(self, m, name, clk, rst, width=32):

        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.width = width

        self.seq = Seq(self.m, self.name, self.clk, self.rst)

        self.lock_reg = self.m.Reg(
            '_'.join(['', self.name, 'lock_reg']), initval=0)
        self.lock_id = self.m.Reg(
            '_'.join(['', self.name, 'lock_id']), self.width, initval=0)

        self.id_map = OrderedDict()
        self.id_map_count = 0

    def lock(self, fsm):
        name = fsm.name
        new_lock_id = self._get_id(name)

        if new_lock_id > 2 ** self.width - 1:
            raise ValueError('too many lock IDs')

        # try
        try_state = fsm.current

        state_cond = fsm.state == fsm.current
        try_cond = vtypes.Not(self.lock_reg)
        fsm_cond = vtypes.Ors(try_cond, self.lock_id == new_lock_id)

        self.seq.If(state_cond, try_cond)(
            self.lock_reg(1),
            self.lock_id(new_lock_id)
        )

        fsm.If(fsm_cond).goto_next()

        # verify
        cond = vtypes.Ands(self.lock_reg, self.lock_id == new_lock_id)
        fsm.If(vtypes.Not(cond)).goto(try_state)  # try again
        fsm.If(cond).goto_next()  # OK

        return 1

    def try_lock(self, fsm):
        name = fsm.name
        new_lock_id = self._get_id(name)

        if new_lock_id > 2 ** self.width - 1:
            raise ValueError('too many lock IDs')

        # try
        try_state = fsm.current

        state_cond = fsm.state == fsm.current
        try_cond = vtypes.Not(self.lock_reg)

        self.seq.If(state_cond, try_cond)(
            self.lock_reg(1),
            self.lock_id(new_lock_id)
        )

        fsm.goto_next()

        # verify
        cond = vtypes.And(self.lock_reg, self.lock_id == new_lock_id)
        result = self.m.TmpReg(initval=0)
        fsm(
            result(cond)
        )
        fsm.goto_next()

        return result

    def unlock(self, fsm):
        name = fsm.name
        new_lock_id = self._get_id(name)

        if new_lock_id > 2 ** self.width - 1:
            raise ValueError('too many lock IDs')

        state_cond = fsm.state == fsm.current

        self.seq.If(state_cond, self.lock_id == new_lock_id)(
            self.lock_reg(0)
        )

        fsm.goto_next()

        return 0

    def _get_id(self, name):
        if name not in self.id_map:
            self.id_map[name] = self.id_map_count
            self.id_map_count += 1

        return self.id_map[name]

    def acquire(self, fsm, blocking=True):
        """ alias of lock() """

        if not isinstance(blocking, (bool, int)):
            raise TypeError('blocking argument must be bool')

        if blocking:
            return self.lock(fsm)

        return self.try_lock(fsm)

    def release(self, fsm):
        """ alias of unlock() """

        return self.unlock(fsm)


class _MutexFunction(object):
    __intrinsics__ = ('lock', 'try_lock', 'unlock')

    def _check_mutex(self, fsm):
        if self.mutex is None:
            self.mutex = Mutex(self.m, '_'.join(
                ['', self.name, 'mutex']), self.clk, self.rst)

    def lock(self, fsm):
        self._check_mutex(fsm)
        return self.mutex.lock(fsm)

    def try_lock(self, fsm):
        self._check_mutex(fsm)
        return self.mutex.try_lock(fsm)

    def unlock(self, fsm):
        self._check_mutex(fsm)
        return self.mutex.unlock(fsm)


class Barrier(object):
    __intrinsics__ = ('wait', )

    def __init__(self, m, name, clk, rst, numparties):

        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.numparties = numparties
        self.width = util.log2(self.numparties) + 1

        self.seq = Seq(self.m, self.name, self.clk, self.rst)

        self.count = self.m.Reg(
            '_'.join(['', self.name, 'barrier_count']), self.width, initval=0)
        self.done = self.m.Reg(
            '_'.join(['', self.name, 'barrier_done']), initval=0)
        self.mutex = Mutex(self.m, '_'.join(
            ['', self.name, 'barrier_mutex']), self.clk, self.rst)

        # reset condition
        self.seq(
            self.done(0)
        )
        self.seq.If(self.count == self.numparties)(
            self.count(0),
            self.done(1)
        )

    def wait(self, fsm):

        self.mutex.lock(fsm)

        state_cond = fsm.state == fsm.current
        self.seq.If(state_cond)(
            self.count.inc()
        )
        fsm.goto_next()

        self.mutex.unlock(fsm)

        fsm.If(self.done).goto_next()

        return 0


class Shared(_MutexFunction):
    __intrinsics__ = ('read', 'write') + _MutexFunction.__intrinsics__

    def __init__(self, value):
        self._value = value
        self.seq = None
        self.mutex = None

    @property
    def value(self):
        return self._value

    def read(self, fsm):
        return self._value

    def write(self, fsm, value, *part):
        if self.seq is None:
            m = fsm.m
            clk = fsm.clk
            rst = fsm.rst
            name = self._value.name
            self.seq = Seq(m, '_'.join(['seq', name]), clk, rst)

        cond = fsm.state == fsm.current

        def getval(v, p):
            if isinstance(p, (tuple, list)):
                return v[p[0], p[1]]
            return v[p]

        if len(part) == 0:
            targ = self._value
        elif len(part) == 1:
            targ = getval(self._value, part[0])
        elif len(part) == 2:
            targ = getval(getval(self._value, part[0]), part[1])
        else:
            raise TypeError('unsupported type')

        self.seq.If(cond)(
            targ(value)
        )

        fsm.goto_next()
        return 0

    def _check_mutex(self, fsm):
        if self.mutex is None:
            m = fsm.m
            clk = fsm.clk
            rst = fsm.rst
            name = self._value.name
            self.mutex = Mutex(m, '_'.join(['', name, 'mutex']), clk, rst)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_axi_slave_readwrite

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [32-1:0] sum;
  reg [32-1:0] myaxi_awaddr;
  reg [8-1:0] myaxi_awlen;
  reg myaxi_awvalid;
  wire myaxi_awready;
  reg [32-1:0] myaxi_wdata;
  reg [4-1:0] myaxi_wstrb;
  reg myaxi_wlast;
  reg myaxi_wvalid;
  wire myaxi_wready;
  reg [32-1:0] myaxi_araddr;
  reg [8-1:0] myaxi_arlen;
  reg myaxi_arvalid;
  wire myaxi_arready;
  wire [32-1:0] myaxi_rdata;
  wire myaxi_rlast;
  wire myaxi_rvalid;
  reg myaxi_rready;
  reg [32-1:0] _axi_awaddr;
  reg [8-1:0] _axi_awlen;
  reg _axi_awvalid;
  wire _axi_awready;
  reg [32-1:0] _axi_wdata;
  reg [4-1:0] _axi_wstrb;
  reg _axi_wlast;
  reg _axi_wvalid;
  wire _axi_wready;
  reg [32-1:0] _axi_araddr;
  reg [8-1:0] _axi_arlen;
  reg _axi_arvalid;
  wire _axi_arready;
  wire [32-1:0] _axi_rdata;
  wire _axi_rlast;
  wire _axi_rvalid;
  wire _axi_rready;
  wire [32-1:0] _tmp_0;
  assign _tmp_0 = _axi_awaddr;

  always @(*) begin
    myaxi_awaddr = _tmp_0;
  end

  wire [8-1:0] _tmp_1;
  assign _tmp_1 = _axi_awlen;

  always @(*) begin
    myaxi_awlen = _tmp_1;
  end

  wire _tmp_2;
  assign _tmp_2 = _axi_awvalid;

  always @(*) begin
    myaxi_awvalid = _tmp_2;
  end

  assign _axi_awready = myaxi_awready;
  wire [32-1:0] _tmp_3;
  assign _tmp_3 = _axi_wdata;

  always @(*) begin
    myaxi_wdata = _tmp_3;
  end

  wire [4-1:0] _tmp_4;
  assign _tmp_4 = _axi_wstrb;

  always @(*) begin
    myaxi_wstrb = _tmp_4;
  end

  wire _tmp_5;
  assign _tmp_5 = _axi_wlast;

  always @(*) begin
    myaxi_wlast = _tmp_5;
  end

  wire _tmp_6;
  assign _tmp_6 = _axi_wvalid;

  always @(*) begin
    myaxi_wvalid = _tmp_6;
  end

  assign _axi_wready = myaxi_wready;
  wire [32-1:0] _tmp_7;
  assign _tmp_7 = _axi_araddr;

  always @(*) begin
    myaxi_araddr = _tmp_7;
  end

  wire [8-1:0] _tmp_8;
  assign _tmp_8 = _axi_arlen;

  always @(*) begin
    myaxi_arlen = _tmp_8;
  end

  wire _tmp_9;
  assign _tmp_9 = _axi_arvalid;

  always @(*) begin
    myaxi_arvalid = _tmp_9;
  end

  assign _axi_arready = myaxi_arready;
  assign _axi_rdata = myaxi_rdata;
  assign _axi_rlast = myaxi_rlast;
  assign _axi_rvalid = myaxi_rvalid;
  wire _tmp_10;
  assign _tmp_10 = _axi_rready;

  always @(*) begin
    myaxi_rready = _tmp_10;
  end

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [32-1:0] rsum;
  reg [9-1:0] _tmp_11;
  reg __axi_cond_0_1;
  reg [9-1:0] _tmp_12;
  reg __axi_cond_1_1;
  assign _axi_rready = (fsm == 1) || (fsm == 3);
  reg [9-1:0] _tmp_13;
  reg __axi_cond_2_1;
  reg [32-1:0] wdata;
  reg _tmp_14;
  reg __axi_cond_3_1;
  reg [9-1:0] _tmp_15;
  reg __axi_cond_4_1;
  reg _tmp_16;
  reg __axi_cond_5_1;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .sum(sum),
    .myaxi_awaddr(myaxi_awaddr),
    .myaxi_awlen(myaxi_awlen),
    .myaxi_awvalid(myaxi_awvalid),
    .myaxi_awready(myaxi_awready),
    .myaxi_wdata(myaxi_wdata),
    .myaxi_wstrb(myaxi_wstrb),
    .myaxi_wlast(myaxi_wlast),
    .myaxi_wvalid(myaxi_wvalid),
    .myaxi_wready(myaxi_wready),
    .myaxi_araddr(myaxi_araddr),
    .myaxi_arlen(myaxi_arlen),
    .myaxi_arvalid(myaxi_arvalid),
    .myaxi_arready(myaxi_arready),
    .myaxi_rdata(myaxi_rdata),
    .myaxi_rlast(myaxi_rlast),
    .myaxi_rvalid(myaxi_rvalid),
    .myaxi_rready(myaxi_rready)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST, sum, myaxi_awaddr, myaxi_awlen, myaxi_awvalid, myaxi_awready, myaxi_wdata, myaxi_wstrb, myaxi_wlast, myaxi_wvalid, myaxi_wready, myaxi_araddr, myaxi_arlen, myaxi_arvalid, myaxi_arready, myaxi_rdata, myaxi_rlast, myaxi_rvalid, myaxi_rready, _axi_awaddr, _axi_awlen, _axi_awvalid, _axi_awready, _axi_wdata, _axi_wstrb, _axi_wlast, _axi_wvalid, _axi_wready, _axi_araddr, _axi_arlen, _axi_arvalid, _axi_arready, _axi_rdata, _axi_rlast, _axi_rvalid, _axi_rready, _tmp_0, _tmp_1, _tmp_2, _tmp_3, _tmp_4, _tmp_5, _tmp_6, _tmp_7, _tmp_8, _tmp_9, _tmp_10, fsm, rsum, _tmp_11, __axi_cond_0_1, _tmp_12, __axi_cond_1_1, _tmp_13, __axi_cond_2_1, wdata, _tmp_14, __axi_cond_3_1, _tmp_15, __axi_cond_4_1, _tmp_16, __axi_cond_5_1);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    _axi_awaddr = 0;
    _axi_awlen = 0;
    _axi_awvalid = 0;
    _axi_wdata = 0;
    _axi_wstrb = 0;
    _axi_wlast = 0;
    _axi_wvalid = 0;
    _axi_araddr = 0;
    _axi_arlen = 0;
    _axi_arvalid = 0;
    fsm = fsm_init;
    rsum = 0;
    _tmp_11 = 0;
    __axi_cond_0_1 = 0;
    _tmp_12 = 0;
    __axi_cond_1_1 = 0;
    _tmp_13 = 0;
    __axi_cond_2_1 = 0;
    wdata = 0;
    _tmp_14 = 0;
    __axi_cond_3_1 = 0;
    _tmp_15 = 0;
    __axi_cond_4_1 = 0;
    _tmp_16 = 0;
    __axi_cond_5_1 = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


  always @(posedge CLK) begin
    if(RST) begin
      _axi_araddr <= 0;
      _axi_arlen <= 0;
      _axi_arvalid <= 0;
      _tmp_11 <= 0;
      __axi_cond_0_1 <= 0;
      _tmp_12 <= 0;
      __axi_cond_1_1 <= 0;
      _axi_awaddr <= 0;
      _axi_awlen <= 0;
      _axi_awvalid <= 0;
      _tmp_13 <= 0;
      __axi_cond_2_1 <= 0;
      _axi_wdata <= 0;
      _axi_wvalid <= 0;
      _axi_wlast <= 0;
      _axi_wstrb <= 0;
      _tmp_14 <= 0;
      __axi_cond_3_1 <= 0;
      _tmp_15 <= 0;
      __axi_cond_4_1 <= 0;
      _tmp_16 <= 0;
      __axi_cond_5_1 <= 0;
    end else begin
      if(__axi_cond_0_1) begin
        _axi_arvalid <= 0;
      end 
      if(__axi_cond_1_1) begin
        _axi_arvalid <= 0;
      end 
      if(__axi_cond_2_1) begin
        _axi_awvalid <= 0;
      end 
      if(__axi_cond_3_1) begin
        _axi_wvalid <= 0;
        _axi_wlast <= 0;
        _tmp_14 <= 0;
      end 
      if(__axi_cond_4_1) begin
        _axi_awvalid <= 0;
      end 
      if(__axi_cond_5_1) begin
        _axi_wvalid <= 0;
        _axi_wlast <= 0;
        _tmp_16 <= 0;
      end 
      if((fsm == 0) && ((_axi_arready || !_axi_arvalid) && (_tmp_11 == 0))) begin
        _axi_araddr <= 1024;
        _axi_arlen <= 63;
        _axi_arvalid <= 1;
        _tmp_11 <= 64;
      end 
      __axi_cond_0_1 <= 1;
      if(_axi_arvalid && !_axi_arready) begin
        _axi_arvalid <= _axi_arvalid;
      end 
      if(_axi_rready && _axi_rvalid && (_tmp_11 > 0)) begin
        _tmp_11 <= _tmp_11 - 1;
      end 
      if((fsm == 2) && ((_axi_arready || !_axi_arvalid) && (_tmp_12 == 0))) begin
        _axi_araddr <= 2048;
        _axi_arlen <= 127;
        _axi_arvalid <= 1;
        _tmp_12 <= 128;
      end 
      __axi_cond_1_1 <= 1;
      if(_axi_arvalid && !_axi_arready) begin
        _axi_arvalid <= _axi_arvalid;
      end 
      if(_axi_rready && _axi_rvalid && (_tmp_12 > 0)) begin
        _tmp_12 <= _tmp_12 - 1;
      end 
      if((fsm == 5) && ((_axi_awready || !_axi_awvalid) && (_tmp_13 == 0))) begin
        _axi_awaddr <= 1024;
        _axi_awlen <= 63;
        _axi_awvalid <= 1;
        _tmp_13 <= 64;
      end 
      if((fsm == 5) && ((_axi_awready || !_axi_awvalid) && (_tmp_13 == 0)) && 0) begin
        _axi_awvalid <= 0;
      end 
      __axi_cond_2_1 <= 1;
      if(_axi_awvalid && !_axi_awready) begin
        _axi_awvalid <= _axi_awvalid;
      end 
      if((fsm == 6) && ((_tmp_13 > 0) && (_axi_wready || !_axi_wvalid) && (_tmp_13 > 0))) begin
        _axi_wdata <= wdata;
        _axi_wvalid <= 1;
        _axi_wlast <= 0;
        _axi_wstrb <= { 4{ 1'd1 } };
        _tmp_13 <= _tmp_13 - 1;
      end 
      if((fsm == 6) && ((_tmp_13 > 0) && (_axi_wready || !_axi_wvalid) && (_tmp_13 > 0)) && (_tmp_13 == 1)) begin
        _axi_wlast <= 1;
        _tmp_14 <= 1;
      end 
      __axi_cond_3_1 <= 1;
      if(_axi_wvalid && !_axi_wready) begin
        _axi_wvalid <= _axi_wvalid;
        _axi_wlast <= _axi_wlast;
        _tmp_14 <= _tmp_14;
      end 
      if((fsm == 7) && ((_axi_awready || !_axi_awvalid) && (_tmp_15 == 0))) begin
        _axi_awaddr <= 1024;
        _axi_awlen <= 127;
        _axi_awvalid <= 1;
        _tmp_15 <= 128;
      end 
      if((fsm == 7) && ((_axi_awready || !_axi_awvalid) && (_tmp_15 == 0)) && 0) begin
        _axi_awvalid <= 0;
      end 
      __axi_cond_4_1 <= 1;
      if(_axi_awvalid && !_axi_awready) begin
        _axi_awvalid <= _axi_awvalid;
      end 
      if((fsm == 8) && ((_tmp_15 > 0) && (_axi_wready || !_axi_wvalid) && (_tmp_15 > 0))) begin
        _axi_wdata <= wdata;
        _axi_wvalid <= 1;
        _axi_wlast <= 0;
        _axi_wstrb <= { 4{ 1'd1 } };
        _tmp_15 <= _tmp_15 - 1;
      end 
      if((fsm == 8) && ((_tmp_15 > 0) && (_axi_wready || !_axi_wvalid) && (_tmp_15 > 0)) && (_tmp_15 == 1)) begin
        _axi_wlast <= 1;
        _tmp_16 <= 1;
      end 
      __axi_cond_5_1 <= 1;
      if(_axi_wvalid && !_axi_wready) begin
        _axi_wvalid <= _axi_wvalid;
        _axi_wlast <= _axi_wlast;
        _tmp_16 <= _tmp_16;
      end 
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;
  localparam fsm_8 = 8;
  localparam fsm_9 = 9;
  localparam fsm_10 = 10;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      rsum <= 0;
      wdata <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          if(_axi_arready || !_axi_arvalid) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if(_axi_rready && _axi_rvalid) begin
            rsum <= rsum + _axi_rdata;
          end 
          if(_axi_rready && _axi_rvalid && _axi_rlast) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          if(_axi_arready || !_axi_arvalid) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(_axi_rready && _axi_rvalid) begin
            rsum <= rsum + _axi_rdata;
          end 
          if(_axi_rready && _axi_rvalid && _axi_rlast) begin
            fsm <= fsm_4;
          end 
        end
        fsm_4: begin
          $display("rsum=%d expected_rsum=%d", rsum, 92064);
          fsm <= fsm_5;
        end
        fsm_5: begin
          if(_axi_awready || !_axi_awvalid) begin
            fsm <= fsm_6;
          end 
        end
        fsm_6: begin
          if((_tmp_13 > 0) && (_axi_wready || !_axi_wvalid)) begin
            wdata <= wdata + 1;
          end 
          if(_tmp_14) begin
            fsm <= fsm_7;
          end 
        end
        fsm_7: begin
          if(_axi_awready || !_axi_awvalid) begin
            fsm <= fsm_8;
          end 
        end
        fsm_8: begin
          if((_tmp_15 > 0) && (_axi_wready || !_axi_wvalid)) begin
            wdata <= wdata + 1;
          end 
          if(_tmp_16) begin
            fsm <= fsm_9;
          end 
        end
        fsm_9: begin
          $display("sum=%d expected_sum=%d", sum, 18336);
          fsm <= fsm_10;
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  output reg [32-1:0] sum,
  input [32-1:0] myaxi_awaddr,
  input [8-1:0] myaxi_awlen,
  input myaxi_awvalid,
  output myaxi_awready,
  input [32-1:0] myaxi_wdata,
  input [4-1:0] myaxi_wstrb,
  input myaxi_wlast,
  input myaxi_wvalid,
  output myaxi_wready,
  input [32-1:0] myaxi_araddr,
  input [8-1:0] myaxi_arlen,
  input myaxi_arvalid,
  output myaxi_arready,
  output reg [32-1:0] myaxi_rdata,
  output reg myaxi_rlast,
  output reg myaxi_rvalid,
  input myaxi_rready
);

  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg [9-1:0] _tmp_0;
  reg [32-1:0] _tmp_1;
  reg _tmp_2;
  reg _tmp_3;
  reg _tmp_4;
  reg _tmp_5;
  assign myaxi_awready = (fsm == 0) && !_tmp_2 && !_tmp_3 && _tmp_4;
  assign myaxi_arready = (fsm == 0) && !_tmp_3 && !_tmp_2 && _tmp_5;
  reg [32-1:0] rdata;
  reg _tmp_6;
  reg _myaxi_cond_0_1;
  assign myaxi_wready = fsm == 100;

  always @(posedge CLK) begin
    if(RST) begin
      _tmp_4 <= 0;
      _tmp_5 <= 0;
      _tmp_2 <= 0;
      _tmp_3 <= 0;
      _tmp_1 <= 0;
      _tmp_0 <= 0;
      myaxi_rdata <= 0;
      myaxi_rvalid <= 0;
      myaxi_rlast <= 0;
      _tmp_6 <= 0;
      _myaxi_cond_0_1 <= 0;
    end else begin
      if(_myaxi_cond_0_1) begin
        myaxi_rvalid <= 0;
        myaxi_rlast <= 0;
        _tmp_6 <= 0;
      end 
      _tmp_4 <= myaxi_awvalid;
      _tmp_5 <= myaxi_arvalid;
      _tmp_2 <= 0;
      _tmp_3 <= 0;
      if(myaxi_awready && myaxi_awvalid) begin
        _tmp_1 <= myaxi_awaddr;
        _tmp_0 <= myaxi_awlen + 1;
        _tmp_2 <= 1;
      end else if(myaxi_arready && myaxi_arvalid) begin
        _tmp_1 <= myaxi_araddr;
        _tmp_0 <= myaxi_arlen + 1;
        _tmp_3 <= 1;
      end 
      if((fsm == 1) && ((_tmp_0 > 0) && (myaxi_rready || !myaxi_rvalid) && (_tmp_0 > 0))) begin
        myaxi_rdata <= rdata;
        myaxi_rvalid <= 1;
        myaxi_rlast <= 0;
        _tmp_0 <= _tmp_0 - 1;
      end 
      if((fsm == 1) && ((_tmp_0 > 0) && (myaxi_rready || !myaxi_rvalid) && (_tmp_0 > 0)) && (_tmp_0 == 1)) begin
        myaxi_rlast <= 1;
        _tmp_6 <= 1;
      end 
      _myaxi_cond_0_1 <= 1;
      if(myaxi_rvalid && !myaxi_rready) begin
        myaxi_rvalid <= myaxi_rvalid;
        myaxi_rlast <= myaxi_rlast;
        _tmp_6 <= _tmp_6;
      end 
      if(myaxi_wready && myaxi_wvalid && (_tmp_0 > 0)) begin
        _tmp_0 <= _tmp_0 - 1;
      end 
    end
  end

  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_100 = 100;
  localparam fsm_101 = 101;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      rdata <= 0;
      sum <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          if(_tmp_3) begin
            rdata <= _tmp_1 >> 2;
          end 
          if(_tmp_2) begin
            fsm <= fsm_100;
          end 
          if(_tmp_3) begin
            fsm <= fsm_1;
          end 
        end
        fsm_1: begin
          if((_tmp_0 > 0) && (myaxi_rready || !myaxi_rvalid)) begin
            rdata <= rdata + 1;
          end 
          if(_tmp_6) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          fsm <= fsm_init;
        end
        fsm_100: begin
          if(myaxi_wready && myaxi_wvalid) begin
            sum <= sum + myaxi_wdata;
          end 
          if(myaxi_wready && myaxi_wvalid && myaxi_wlast) begin
            fsm <= fsm_101;
          end 
        end
        fsm_101: begin
          fsm <= fsm_init;
        end
      endcase
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_axi_slave_readwrite.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)
